# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with classes and methods that manage the AWS ApiGateway at high level."""

from typing import Dict
from scar.providers.aws import GenericClient
import scar.logger as logger


class APIGateway(GenericClient):
    """Manage the calls to the ApiGateway client."""

    def __init__(self, resources_info: Dict):
        super().__init__(resources_info.get('api_gateway', {}))
        self.resources_info = resources_info
        self.api = self.resources_info.get('api_gateway', {})

    def _get_common_args(self) -> Dict:
        return {'restApiId' : self.api.get('id', ''),
                'resourceId' : self.api.get('resource_id', ''),
                'httpMethod' : self.api.get('http_method', '')}

    def _get_method_args(self) -> Dict:
        args = self._get_common_args()
        args.update(self.api.get('method', {}))
        return args

    def _get_integration_args(self) -> Dict:
        integration_args = self.api.get('integration', {})
        uri_args = {'api_region': self.api.get('region', ''),
                    'lambda_region': self.resources_info.get('lambda', {}).get('region', ''),
                    'account_id': self.resources_info.get('iam', {}).get('account_id', ''),
                    'function_name': self.resources_info.get('lambda', {}).get('name', '')}
        integration_args['uri'] = integration_args['uri'].format(**uri_args)
        args = self._get_common_args()
        args.update(integration_args)
        return args

    def _get_resource_id(self) -> str:
        res_id = ""
        resources_info = self.client.get_resources(self.api.get('id', ''))
        for resource in resources_info['items']:
            if resource['path'] == '/':
                res_id = resource['id']
                break
        return res_id

    def _set_api_gateway_id(self, api_info: Dict) -> None:
        self.api['id'] = api_info.get('id', '')
        # We store the parameter in the lambda configuration that
        # is going to be uploaded to the Lambda service
        self.resources_info['lambda']['environment']['Variables']['API_GATEWAY_ID'] = api_info.get('id', '')

    def _set_resource_info_id(self, resource_info: Dict) -> None:
        self.api['resource_id'] = resource_info.get('id', '')

    def _get_endpoint(self) -> str:
        endpoint_args = {'api_id': self.api.get('id', ''),
                         'api_region': self.api.get('region', ''),
                         'stage_name': self.api.get('stage_name', '')}
        return self.api.get('endpoint', '').format(**endpoint_args)

    def create_api_gateway(self) -> None:
        """Creates an Api Gateway endpoint."""
        api_info = self.client.create_rest_api(self.api.get('name', ''))
        self._set_api_gateway_id(api_info)
        resource_info = self.client.create_resource(self.api.get('id', ''),
                                                    self._get_resource_id(),
                                                    self.api.get('path_part', ''))
        self._set_resource_info_id(resource_info)
        self.client.create_method(**self._get_method_args())
        self.client.set_integration(**self._get_integration_args())
        self.client.create_deployment(self.api.get('id', ''), self.api.get('stage_name', ''))
        logger.info(f'API Gateway endpoint: {self._get_endpoint()}')

    def delete_api_gateway(self) -> None:
        """Deletes an Api Gateway endpoint."""
        return self.client.delete_rest_api(self.resources_info['lambda']['environment']['Variables']['API_GATEWAY_ID'])



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with the class necessary to manage the
Batch creation, deletion and configuration."""

from typing import Dict
from scar.providers.aws.clients import BotoClient
import scar.logger as logger
from scar.exceptions import exception


class BatchClient(BotoClient):
    """A low-level client representing aws batchClient.
    DOC_URL: http://boto3.readthedocs.io/en/latest/reference/services/batch.html"""

    # Parameter used by the parent to create the appropriate boto3 client
    _BOTO_CLIENT_NAME = 'batch'

    @exception(logger)
    def create_compute_environment(self, **kwargs: Dict) -> Dict:
        """Creates a new compute environment."""
        logger.debug("Creating compute environment.")
        return self.client.create_compute_environment(**kwargs)

    @exception(logger)
    def create_job_queue(self, **kwargs: Dict) -> Dict:
        """Creates a new job queue."""
        logger.debug("Creating job queue.")
        return self.client.create_job_queue(**kwargs)

    @exception(logger)
    def register_job_definition(self, **kwargs: Dict) -> Dict:
        """Registers a new job definition."""
        logger.debug("Registering job definition.")
        return self.client.register_job_definition(**kwargs)

    @exception(logger)
    def describe_compute_environments(self, **kwargs: Dict) -> Dict:
        """Describes compute environments."""
        logger.debug("Describing Compute Environment.")
        return self.client.describe_compute_environments(**kwargs)

    @exception(logger)
    def describe_job_queues(self, **kwargs: Dict) -> Dict:
        """Describe a new job queue."""
        logger.debug("Describing job queue.")
        return self.client.describe_job_queues(**kwargs)

    @exception(logger)
    def describe_job_definitions(self, **kwargs: Dict) -> Dict:
        """Describes a list of job definitions."""
        logger.debug("Describing job definition.")
        return self.client.describe_job_definitions(**kwargs)

    @exception(logger)
    def deregister_job_definition(self, **kwargs: Dict) -> Dict:
        """Deregisters an AWS Batch job definition."""
        logger.debug("Deleting job definition.")
        return self.client.deregister_job_definition(**kwargs)

    @exception(logger)
    def update_job_queue(self, **kwargs: Dict) -> Dict:
        """Updates a job queue."""
        logger.debug("Updating job queue.")
        return self.client.update_job_queue(**kwargs)

    @exception(logger)
    def delete_job_queue(self, **kwargs: Dict) -> Dict:
        """Deletes a job queue."""
        logger.debug("Deleting job queue.")
        return self.client.delete_job_queue(**kwargs)

    @exception(logger)
    def update_compute_environment(self, **kwargs: Dict) -> Dict:
        """Updates a compute environment."""
        logger.debug("Updating compute environment.")
        return self.client.update_compute_environment(**kwargs)

    @exception(logger)
    def delete_compute_environment(self, **kwargs: Dict) -> Dict:
        """Deletes a compute environment."""
        logger.debug("Deleting compute environment.")
        return self.client.delete_compute_environment(**kwargs)

    @exception(logger)
    def describe_jobs(self, **kwargs: Dict) -> Dict:
        """Describes a batch job."""
        logger.debug("Describing a job.")
        return self.client.describe_jobs(**kwargs)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests

def call_http_endpoint(url, **kwargs):
    """
    Does a 'GET' or 'PUT' request if the parameter 'data' exists or not respectively

    :param url: URL for the request.        
    :param data: (optional) Dictionary (will be form-encoded), bytes, or file-like object to send in the body of the request.
    :param headers: (optional) Dictionary of HTTP Headers to send with the request.
    :param parameters: (optional) Dictionary or bytes to be sent in the query string.
    """
    if ('data' in kwargs and kwargs['data']) or ('json' in kwargs and kwargs['json']):
        response = requests.post(url, **kwargs)
    else:
        response = requests.get(url, **kwargs)
    return response

def get_file(url):
    response = requests.get(url)
    if response:
        return response.content
    


# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os

log_folder_name = ".scar"
log_file_folder = os.path.join(os.path.expanduser("~"), log_folder_name)
if 'SCAR_LOG_PATH' in os.environ:
    log_file_folder = os.environ['SCAR_LOG_PATH']
log_file_name = "scar.log"
log_file_path = os.path.join(log_file_folder, log_file_name)

# Create scar config dir
if not os.path.isdir(log_file_folder):
    os.makedirs(log_file_folder, exist_ok=True)

loglevel = logging.INFO
FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
logging.basicConfig(filename=log_file_path, level=loglevel, format=FORMAT)

def init_execution_trace():
    logging.info('----------------------------------------------------')
    logging.info('SCAR execution started')
    
def end_execution_trace():
    logging.info('SCAR execution finished')
    logging.info('----------------------------------------------------')     
        
def end_execution_trace_with_errors():
    logging.info('SCAR execution finished with errors')
    logging.info('----------------------------------------------------')

def debug(cli_msg, log_msg=None):
    if loglevel == logging.DEBUG:
        print(cli_msg)
    logging.debug(log_msg) if log_msg else logging.debug(cli_msg)

def info(cli_msg=None, log_msg=None):
    if cli_msg and loglevel == logging.INFO:
        print(cli_msg)
    logging.info(log_msg) if log_msg else logging.info(cli_msg)

def warning(cli_msg, log_msg=None):
    print(cli_msg)
    logging.warning(log_msg) if log_msg else logging.warning(cli_msg)

def error(cli_msg, log_msg=None):
    if log_msg:
        print(log_msg)
        logging.error(log_msg)
    else:
        print(cli_msg)
        logging.error(cli_msg)
        
def exception(msg):
    logging.exception(msg)        

def log_exception(error_msg, exception):
    error(error_msg, error_msg + ": {0}".format(exception))

def print_json(value):
    print(json.dumps(value))

def info_json(cli_msg, log_msg=None):
    print_json(cli_msg)
    logging.info(log_msg) if log_msg else logging.info(cli_msg)

def warning_json(cli_msg, log_msg=None):
    print_json(cli_msg)
    logging.warning(log_msg) if log_msg else logging.warning(cli_msg)

def error_json(cli_msg, log_msg=None):
    print_json(cli_msg)
    logging.error(log_msg) if log_msg else logging.error(cli_msg)          



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with the generic classes and methods used in other aws classes"""

from typing import Dict
from scar.utils import lazy_property
from scar.providers.aws.clients.apigateway import APIGatewayClient
from scar.providers.aws.clients.batchfunction import BatchClient
from scar.providers.aws.clients.cloudwatchlogs import CloudWatchLogsClient
from scar.providers.aws.clients.iam import IAMClient
from scar.providers.aws.clients.lambdafunction import LambdaClient
from scar.providers.aws.clients.resourcegroups import ResourceGroupsClient
from scar.providers.aws.clients.s3 import S3Client
from scar.providers.aws.clients.ec2 import EC2Client


class GenericClient():
    """Class in charge of creating the boto clients."""

    # pylint: disable=too-few-public-methods

    _CLIENTS = {'APIGATEWAY': APIGatewayClient,
                'BATCH': BatchClient,
                'CLOUDWATCHLOGS': CloudWatchLogsClient,
                'IAM': IAMClient,
                'LAMBDA': LambdaClient,
                'RESOURCEGROUPS': ResourceGroupsClient,
                'S3': S3Client,
                'LAUNCHTEMPLATES': EC2Client}

    def __init__(self, resource_info: Dict =None):
        self.properties = {}
        if resource_info:
            region = resource_info.get('region')
            if region:
                self.properties['client'] = {'region_name': region}
            session = resource_info.get('boto_profile')
            if session:
                self.properties['session'] = {'profile_name': session}        

    @lazy_property
    def client(self):
        """Returns the required boto client based on the implementing class name."""
        client = self._CLIENTS[self.__class__.__name__.upper()](self.properties)
        return client



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Dict
from enum import Enum
from tabulate import tabulate
import scar.logger as logger
from scar.utils import StrUtils
from requests import Response


class OutputType(Enum):
    PLAIN_TEXT = 1
    JSON = 2
    VERBOSE = 3
    BINARY = 4

def parse_http_response(response: Response, resources_info: Dict, scar_info: Dict) -> None:
    '''Process the response generated by an API Gateway invocation.'''
    output_type = scar_info.get('cli_output')
    function_name = resources_info.get('lambda').get('name')
    asynch = resources_info.get('lambda').get('asynchronous')
    text_message = ""
    if response.ok:
        if output_type == OutputType.BINARY.value:
            output_file = scar_info.get('output_file', '')
            with open(output_file, "wb") as out:
                out.write(StrUtils.decode_base64(response.text))
            text_message = f"Output saved in file '{output_file}'"
        else:
            text_message = f"Request Id: {response.headers['amz-lambda-request-id']}"
            if asynch:
                text_message += f"\nFunction '{function_name}' launched correctly"
            else:
                text_message += f"\nLog Group Name: {response.headers['amz-log-group-name']}\n"
                text_message += f"Log Stream Name: {response.headers['amz-log-stream-name']}\n"
                text_message += StrUtils.base64_to_utf8_string(response.text)
    else:
        if asynch and response.status_code == 502:
            text_message = f"Function '{function_name}' launched successfully."
        else:
            error = json.loads(response.text)
            if 'message' in error:
                text_message = f"Error ({response.reason}): {error['message']}"
            else:
                text_message = f"Error ({response.reason}): {error['exception']}"
    logger.info(text_message)


def _print_generic_response(response, output_type, aws_output, text_message=None, json_output=None, verbose_output=None, output_file=None):
    if output_type == OutputType.BINARY.value:
        with open(output_file, "wb") as out:
            out.write(StrUtils.decode_base64(response['Payload']['body']))
    elif output_type == OutputType.PLAIN_TEXT.value:
        output = text_message
        logger.info(output)
    else:
        if output_type == OutputType.JSON.value:
            output = json_output if json_output else {aws_output :
                                                      {'RequestId' : response['ResponseMetadata']['RequestId'],
                                                       'HTTPStatusCode' : response['ResponseMetadata']['HTTPStatusCode']}}
        elif output_type == OutputType.VERBOSE.value:
            output = verbose_output if verbose_output else {aws_output : response}
        logger.info_json(output)


def parse_lambda_function_creation_response(response, output_type, access_key):
    if response:
        aws_output = 'LambdaOutput'
        text_message = f"Function '{response['FunctionName']}' successfully created."
        json_message = {aws_output : {
                            'AccessKey' : access_key,
                            'FunctionArn' : response['FunctionArn'],
                            'Timeout' : response['Timeout'],
                            'MemorySize' : response['MemorySize'],
                            'FunctionName' : response['FunctionName']}
                        }
        _print_generic_response(response, output_type, aws_output, text_message, json_output=json_message)


def parse_log_group_creation_response(response, log_group_name, output_type):
    if response:
        text_message = f"Log group '{log_group_name}' successfully created."
        _print_generic_response(response, output_type, 'CloudWatchOutput', text_message)


def parse_delete_function_response(response, function_name, output_type):
    if response:
        text_message = f"Function '{function_name}' successfully deleted."
        _print_generic_response(response, output_type, 'LambdaOutput', text_message)


def parse_delete_log_response(response, log_group_name, output_type):
    if response:
        text_message = f"Log group '{log_group_name}' successfully deleted."
        _print_generic_response(response, output_type, 'CloudWatchOutput', text_message)


def parse_delete_api_response(response, api_id, output_type):
    if response:
        text_message = f"API Endpoint '{api_id}' successfully deleted."
        _print_generic_response(response, output_type, 'APIGateway', text_message)


def parse_ls_response(aws_resources: Dict, output_type: int) -> None:
    aws_output = 'Functions'
    result = []
    text_message = ""
    if output_type == OutputType.VERBOSE.value:
        result = aws_resources
    else:
        for resources_info in aws_resources:
            result.append(_parse_lambda_function_info(resources_info))
        text_message = _get_table(result)
    json_message = {aws_output: result}
    _print_generic_response('', output_type, aws_output, text_message, json_output=json_message, verbose_output=json_message)


def _parse_lambda_function_info(resources_info: Dict) -> Dict:
    api_gateway = resources_info.get('lambda').get('environment').get('Variables').get('API_GATEWAY_ID', "-")
    if api_gateway != '-':
        stage_name = resources_info.get('api_gateway').get('stage_name')
        region = resources_info.get('api_gateway').get('region')
        api_gateway = f"https://{api_gateway}.execute-api.{region}.amazonaws.com/{stage_name}/launch"
    return {'Name': resources_info.get('lambda').get('name', "-"),
            'Memory': resources_info.get('lambda').get('memory', "-"),
            'Timeout': resources_info.get('lambda').get('timeout', "-"),
            'Image_id': resources_info.get('lambda').get('environment').get('Variables').get('IMAGE_ID', "-"),
            'Api_gateway': api_gateway,
            'Sup_version': resources_info.get('lambda').get('supervisor').get('version', '-')}


def _get_table(functions_info):
    headers = ['NAME', 'MEMORY', 'TIME', 'IMAGE_ID', 'API_URL', 'SUPERVISOR_VERSION']
    table = []
    for function in functions_info:
        table.append([function['Name'],
                      function['Memory'],
                      function['Timeout'],
                      function['Image_id'],
                      function['Api_gateway'],
                      function['Sup_version']])
    return tabulate(table, headers)


def _parse_error_invocation_response(response, function_name):
    if response:
        if "Task timed out" in response['Payload']:
            # Find the timeout time
            message = StrUtils.find_expression(str(response['Payload']), '(Task timed out .* seconds)')
            # Modify the error message to ease the error readability
            error_msg = message.replace("Task", "Function '%s'" % function_name)
            error_log = f"Error in function response: {error_msg}"
        else:
            error_msg = "Error in function response."
            error_log = f"Error in function response: {response['Payload']}"
        logger.error(error_msg, error_log)


def _parse_payload(value):
    if 'Payload' in value and value['Payload']:
        payload = value['Payload'].read()
        if len(payload) > 0:
            value['Payload'] = json.loads(payload.decode("utf-8"))


def _parse_asynchronous_invocation_response(response, output_type, function_name):
    if response:
        aws_output = 'LambdaOutput'
        text_message = f"Request Id: {response['ResponseMetadata']['RequestId']}\n"
        text_message += f"Function '{function_name}' launched correctly"
        json_message = {aws_output : {'StatusCode' : response['StatusCode'],
                                       'RequestId' : response['ResponseMetadata']['RequestId']}}
        _print_generic_response(response, output_type, aws_output, text_message, json_output=json_message)


def _parse_requestresponse_invocation_response(**kwargs):
    if kwargs['Response']:
        response = kwargs['Response']
        aws_output = 'LambdaOutput'
        log_group_name = response['Payload']['headers']['amz-log-group-name']
        log_stream_name = response['Payload']['headers']['amz-log-stream-name']
        request_id = response['ResponseMetadata']['RequestId']
        if "exception" in response['Payload']['body']:
            body = ("ERROR launching udocker container: \n "
                    f"{json.loads(response['Payload']['body'])['exception']}")
        else:
            body = StrUtils.base64_to_utf8_string(response['Payload']['body'])

        text_message = (f"Request Id: {request_id}\n"
                        f"Log Group Name: {log_group_name}\n"
                        f"Log Stream Name: {log_stream_name}\n")
        text_message += body

        json_message = {aws_output : {'StatusCode' : response['StatusCode'],
                                      'Payload' : body,
                                      'LogGroupName' : log_group_name,
                                      'LogStreamName' : log_stream_name,
                                      'RequestId' : request_id}}
        if 'OutputFile' in kwargs and kwargs['OutputFile']:
            _print_generic_response(response, kwargs['OutputType'], aws_output, text_message, json_output=json_message, output_file=kwargs['OutputFile'])
        else:
            _print_generic_response(response, kwargs['OutputType'], aws_output, text_message, json_output=json_message)


def _parse_base64_response_values(value):
    value['LogResult'] = StrUtils.base64_to_utf8_string(value['LogResult'])
    value['ResponseMetadata']['HTTPHeaders']['x-amz-log-result'] = \
        StrUtils.base64_to_utf8_string(value['ResponseMetadata']['HTTPHeaders']['x-amz-log-result'])


def parse_invocation_response(**kwargs):
    # Decode and parse the payload
    _parse_payload(kwargs['Response'])
    if "FunctionError" in kwargs['Response']:
        _parse_error_invocation_response(kwargs['Response'], kwargs['FunctionName'])
    if kwargs['IsAsynchronous']:
        _parse_asynchronous_invocation_response(kwargs['Response'], kwargs['OutputType'], kwargs['FunctionName'])
    else:
        # Transform the base64 encoded results to something legible
        _parse_base64_response_values(kwargs['Response'])
        # Extract log_group_name and log_stream_name from the payload
        _parse_requestresponse_invocation_response(**kwargs)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with the class necessary to manage the
Lambda function and layers creation, deletion and configuration."""

from typing import Dict, List, Optional
from scar.providers.aws.clients import BotoClient
import scar.exceptions as excp
import scar.logger as logger
from scar.utils import StrUtils


class LambdaClient(BotoClient):
    """A low-level client representing aws LambdaClient.
    DOC_URL: https://boto3.readthedocs.io/en/latest/reference/services/lambda.html"""

    # Parameter used by the parent to create the appropriate boto3 client
    _BOTO_CLIENT_NAME = 'lambda'

    @excp.exception(logger)
    def create_function(self, **kwargs: Dict) -> Dict:
        """Creates a new Lambda function."""
        logger.debug("Creating lambda function.")
        return self.client.create_function(**kwargs)

    def get_function_configuration(self, function_name_or_arn: str) -> Dict:
        """Returns the configuration information
        of the Lambda function."""
        function_info = self.client.get_function_configuration(FunctionName=function_name_or_arn)
        # Add supervisor version
        function_info['SupervisorVersion'] = self.get_supervisor_version(function_info)
        return function_info

    def get_function(self, function_name_or_arn: str) -> Dict:
        """Returns the information of the Lambda function with a link to
        download the deployment package that's valid for 10 minutes."""
        function_info = self.client.get_function(FunctionName=function_name_or_arn)
        # Add supervisor version
        function_info['SupervisorVersion'] = self.get_supervisor_version(function_info)
        return function_info

    @excp.exception(logger)
    def get_supervisor_version(self, function_info):
        version = '-'
        # Add supervisor version
        layers = function_info.get('Layers', [])
        for layer in layers:
            layer_arn = layer.get('Arn', '')
            # {'Arn': 'arn:aws:lambda:us-east-1:974349055189:layer:faas-supervisor:1'}
            arn_fields = layer_arn.split(":")
            if arn_fields[-2] == 'faas-supervisor':
                layer_info = self.client.get_layer_version(LayerName='faas-supervisor',
                                                           VersionNumber=int(arn_fields[-1]))
                version = layer_info.get('Description', '-')
        return version

    @excp.exception(logger)
    def update_function_configuration(self, **kwargs: Dict) -> Dict:
        """Updates the configuration parameters for the specified
        Lambda function by using the values provided in the request."""
        # Retrieve the global variables already defined
        return self.client.update_function_configuration(**kwargs)

    @excp.exception(logger)
    def list_functions(self, next_token: Optional[str]=None) -> List:
        """Returns a list of your Lambda functions."""
        logger.debug("Listing lambda functions.")
        functions = []
        kwargs = {}
        if next_token:
            kwargs['Marker'] = next_token
        functions_info = self.client.list_functions(**kwargs)
        if 'Functions' in functions_info and functions_info['Functions']:
            functions.extend(functions_info['Functions'])
        if 'NextMarker' in functions:
            functions.extend(self.list_layers(next_token=functions_info['NextMarker']))
        return functions

    @excp.exception(logger)
    def list_layers(self, next_token: Optional[str]=None) -> List:
        """Lists function layers and shows information
        about the latest version of each."""
        logger.debug("Listing lambda layers.")
        layers = []
        kwargs = {}
        if next_token:
            kwargs['Marker'] = next_token
        layers_info = self.client.list_layers(**kwargs)
        if 'Layers' in layers_info and layers_info['Layers']:
            layers.extend(layers_info['Layers'])
        if 'NextMarker' in layers_info:
            layers.extend(self.list_layers(next_token=layers_info['NextMarker']))
        return layers

    @excp.exception(logger)
    def list_layer_versions(self, layer_name: str, next_token: Optional[str]=None) -> str:
        """Lists the versions of an AWS Lambda layer."""
        logger.debug(f'Listing versions of lambda layer "{layer_name}".')
        versions = []
        kwargs = {'LayerName': layer_name}
        if next_token:
            kwargs['Marker'] = next_token
        layer_versions_info = self.client.list_layer_versions(**kwargs)
        if 'LayerVersions' in layer_versions_info and layer_versions_info['LayerVersions']:
            versions.extend(layer_versions_info['LayerVersions'])
        if 'NextMarker' in layer_versions_info:
            versions.extend(self.list_layer_versions(layer_name, next_token=layer_versions_info['NextMarker']))
        return versions

    @excp.exception(logger)
    def delete_function(self, function_name: str) -> Dict:
        """Deletes the specified Lambda
        function code and configuration."""
        # Delete the lambda function
        return self.client.delete_function(FunctionName=function_name)

    @excp.exception(logger)
    def invoke_function(self, **kwargs: Dict) -> Dict:
        """Invokes a specific Lambda function."""
        return self.client.invoke(**kwargs)

    @excp.exception(logger)
    def add_invocation_permission(self, **kwargs: Dict) -> Dict:
        """Adds a permission to the resource policy associated
        with the specified AWS Lambda function."""
        kwargs['StatementId'] = StrUtils.get_random_uuid4_str()
        kwargs['Action'] = "lambda:InvokeFunction"
        return self.client.add_permission(**kwargs)

    def publish_layer_version(self, **kwargs: Dict) -> Dict:
        """Creates a function layer from a ZIP archive."""
        logger.debug("Publishing lambda layer.")
        return self.client.publish_layer_version(**kwargs)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with the class necessary to manage the
API Gateway creation, deletion and configuration."""

import time
from typing import Dict
from botocore.exceptions import ClientError
from scar.providers.aws.clients import BotoClient
from scar.exceptions import exception, NotExistentApiGatewayWarning
import scar.logger as logger


class APIGatewayClient(BotoClient):
    """A low-level client representing Amazon API Gateway.
    DOC_URL: https://boto3.readthedocs.io/en/latest/reference/services/apigateway.html
    """

    # Parameter used by the parent to create the appropriate boto3 client
    _BOTO_CLIENT_NAME = 'apigateway'
    _ENDPOINT_CONFIGURATION = {'types': ['REGIONAL']}
    _API_DESCRIPTION = "API created automatically with SCAR"
    _MAX_NUMBER_OF_RETRIES = 5
    _WAIT_BETWEEN_RETIRES = 5

    @exception(logger)
    def create_rest_api(self, api_name: str, count: int = _MAX_NUMBER_OF_RETRIES) -> Dict:
        """Creates a new RestApi resource."""
        try:
            api_args = {'name': api_name,
                        'description': self._API_DESCRIPTION,
                        'endpointConfiguration': self._ENDPOINT_CONFIGURATION}
            return self.client.create_rest_api(**api_args)
        except ClientError as cerr:
            if (cerr.response['Error']['Code'] == 'TooManyRequestsException') \
                and (self._MAX_NUMBER_OF_RETRIES > 0):
                time.sleep(self._WAIT_BETWEEN_RETIRES)
                return self.create_rest_api(api_name, count - 1)
            raise cerr

    @exception(logger)
    def get_resources(self, api_id: str) -> Dict:
        """Lists information about a collection of Resource resources."""
        return self.client.get_resources(restApiId=api_id)

    @exception(logger)
    def create_resource(self, api_id: str, parent_id: str, path_part: str) -> Dict:
        """Creates a new RestApi resource."""
        api_args = {'restApiId': api_id,
                    'parentId': parent_id,
                    'pathPart': path_part}
        return self.client.create_resource(**api_args)

    @exception(logger)
    def create_method(self, **kwargs: Dict) -> Dict:
        """Add a method to an existing Resource resource."""
        return self.client.put_method(**kwargs)

    @exception(logger)
    def set_integration(self, **kwargs: Dict) -> Dict:
        """Sets up a method's integration.
        See https://docs.aws.amazon.com/apigateway/latest/
            developerguide/set-up-lambda-proxy-integrations.html"""
        return self.client.put_integration(**kwargs)

    @exception(logger)
    def create_deployment(self, api_id: str, stage_name: str) -> Dict:
        """Creates a Deployment resource, which makes a
        specified RestApi callable over the internet."""
        api_args = {'restApiId': api_id,
                    'stageName': stage_name}
        return self.client.create_deployment(**api_args)

    @exception(logger)
    def delete_rest_api(self, api_id: str, count: int = _MAX_NUMBER_OF_RETRIES) -> Dict:
        """Deletes the specified API."""
        try:
            return self.client.delete_rest_api(restApiId=api_id)
        except ClientError as cerr:
            if (cerr.response['Error']['Code'] == 'TooManyRequestsException') \
                and (self._MAX_NUMBER_OF_RETRIES > 0):
                time.sleep(self._WAIT_BETWEEN_RETIRES)
                return self.delete_rest_api(api_id, count - 1)
            elif cerr.response['Error']['Code'] == 'NotFoundException':
                raise NotExistentApiGatewayWarning(restApiId=api_id)
            raise cerr



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with methods to manage EC2 launch templates.

Generic response from boto3:
{'LaunchTemplates': [{'CreateTime': datetime.datetime(2019, 6, 18, 10, 9, 46, tzinfo=tzutc()),
              'CreatedBy': 'arn:aws:iam::XXX:user/test',
              'DefaultVersionNumber': 1,
              'LatestVersionNumber': 1,
              'LaunchTemplateId': 'lt-052fafdsf8bf047fa9',
              'LaunchTemplateName': 'test',
              'Tags': [{'Key': 'supervisor_version', 'Value': '121'}]}],
 'ResponseMetadata': {'HTTPHeaders': {'content-type': 'text/xml;charset=UTF-8',
                              'date': 'Tue, 18 Jun 2019 10:16:50 GMT',
                              'server': 'AmazonEC2',
                              'transfer-encoding': 'chunked',
                              'vary': 'accept-encoding'},
              'HTTPStatusCode': 200,
              'RequestId': 'XXX',
              'RetryAttempts': 0}}"""

from typing import Dict
from string import Template
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from scar.providers.aws import GenericClient
from scar.utils import SupervisorUtils, StrUtils
import scar.exceptions as excp
import scar.logger as logger


class LaunchTemplates(GenericClient):
    """Class to manage the creation and update of launch templates."""

    # Script to download 'faas-supervisor'
    _LAUNCH_TEMPLATE_SCRIPT = Template(
        '#!/bin/bash\n'
        'mkdir -p /opt/faas-supervisor/bin\n'
        'curl $supervisor_binary_url -L -o /opt/faas-supervisor/bin/supervisor\n'
        'chmod +x /opt/faas-supervisor/bin/supervisor\n'
    )

    def __init__(self, resources_info: Dict):
        super().__init__(resources_info.get('batch'))
        self.supervisor_version = resources_info.get('lambda').get('supervisor').get('version')
        self.template_name = resources_info.get('batch').get('compute_resources').get('launch_template_name')

    @excp.exception(logger)
    def _is_supervisor_created(self) -> bool:
        """Checks if 'faas-supervisor' launch template is created"""
        params = {'Filters': [
            {'Name': 'launch-template-name',
             'Values': [self.template_name]}
        ]}
        response = self.client.describe_launch_templates(params)
        return ('LaunchTemplates' in response and
                response['LaunchTemplates'])

    @excp.exception(logger)
    def _is_supervisor_version_created(self) -> int:
        """Checks if the supervisor version specified is created.
        Returns the Launch Template version or -1 if it does not exists"""
        response = self.client.describe_launch_template_versions(
            {'LaunchTemplateName': self.template_name})
        versions = response['LaunchTemplateVersions']
        # Get ALL versions
        while ('NextToken' in response and response['NextToken']):
            response = self.client.describe_launch_template_versions(
                {'LaunchTemplateName': self.template_name,
                 'NextToken': response['NextToken']})
            versions.extend(response['LaunchTemplateVersions'])

        for version in versions:
            if 'VersionDescription' in version:
                if version['VersionDescription'] == self.supervisor_version:
                    return version['VersionNumber']
        return -1

    @excp.exception(logger)
    def _create_supervisor_user_data(self) -> str:
        """Returns the user_data with the script required for downloading
        the specified version of faas-supervisor in mime-multipart format
        and encoded in base64

        Generic mime-multipart file:
        Content-Type: multipart/mixed; boundary="===============3595946014116037730=="
        MIME-Version: 1.0

        --===============3595946014116037730==
        Content-Type: text/x-shellscript; charset="us-ascii"
        MIME-Version: 1.0
        Content-Transfer-Encoding: 7bit

        #!/bin/bash
        mkdir -p /opt/faas-supervisor/bin
        curl https://github.com/grycap/faas-supervisor/releases/download/1.0.11/supervisor -L -o /opt/faas-supervisor/bin/supervisor
        chmod +x /opt/faas-supervisor/bin/supervisor
        --===============3595946014116037730==--"""
        multipart = MIMEMultipart()
        url = SupervisorUtils.get_supervisor_binary_url(self.supervisor_version)
        script = self._LAUNCH_TEMPLATE_SCRIPT.substitute(
            supervisor_binary_url=url)
        content = MIMEText(script, 'x-shellscript')
        multipart.attach(content)
        return StrUtils.utf8_to_base64_string(str(multipart))

    @excp.exception(logger)
    def get_launch_template_version(self) -> int:
        """Return the launch template version of the specified version of
        'faas-supervisor'. If it does not exists creates a new one."""
        if self._is_supervisor_created():
            is_created = self._is_supervisor_version_created()
            if is_created is not -1:
                logger.info(f"Using existent '{self.template_name}' launch template.")
                return is_created
            else:
                logger.info((f"Creating launch template version with 'faas-supervisor' "
                             f"version '{self.supervisor_version}'."))
                user_data = self._create_supervisor_user_data()
                response = self.client.create_launch_template_version(
                    self.template_name,
                    self.supervisor_version,
                    {'UserData': user_data})
                return response['LaunchTemplateVersion']['VersionNumber']
        else:
            logger.info((f"Creating launch template with 'faas-supervisor' "
                         f"version '{self.supervisor_version}'."))
            user_data = self._create_supervisor_user_data()
            response = self.client.create_launch_template(
                self.template_name,
                self.supervisor_version,
                {'UserData': user_data})
            return response['LaunchTemplate']['LatestVersionNumber']



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with classes and methods used to manage the AWS tools."""

import os
from typing import Dict
from copy import deepcopy
from scar.cmdtemplate import Commands
from scar.providers.aws.apigateway import APIGateway
from scar.providers.aws.batchfunction import Batch
from scar.providers.aws.cloudwatchlogs import CloudWatchLogs
from scar.providers.aws.iam import IAM
from scar.providers.aws.lambdafunction import Lambda
# from scar.providers.aws.properties import AwsProperties, ScarProperties
from scar.providers.aws.resourcegroups import ResourceGroups
from scar.providers.aws.s3 import S3, get_bucket_and_folders
from scar.providers.aws.validators import AWSValidator
import scar.exceptions as excp
import scar.logger as logger
import scar.providers.aws.response as response_parser
from scar.utils import StrUtils, FileUtils, SupervisorUtils

_ACCOUNT_ID_REGEX = r'\d{12}'


def _get_owner(resources_info: Dict):
    return IAM(resources_info).get_user_name_or_id()


def _check_function_defined(resources_info: Dict):
    if Lambda(resources_info).find_function():
        raise excp.FunctionExistsError(function_name=resources_info.get('lambda', {}).get('name', ''))


def _check_function_not_defined(resources_info: Dict):
    if not Lambda(resources_info).find_function():
        raise excp.FunctionNotFoundError(function_name=resources_info.get('lambda', {}).get('name', ''))


def _choose_function(aws_resources: Dict) -> int:
    function_names = [resources_info.get('lambda').get('name') for resources_info in aws_resources]
    print("Please choose a function:")
    print("0) Apply to all")
    for idx, element in enumerate(function_names):
        print(f"{idx+1}) {element}")
    i = input("Enter number: ")
    if 0 <= int(i) <= len(function_names):
        return int(i) - 1
    return None


def _get_all_functions(resources_info: Dict):
    arn_list = ResourceGroups(resources_info).get_resource_arn_list(IAM(resources_info).get_user_name_or_id())
    return Lambda(resources_info).get_all_functions(arn_list)

def _check_preheat_function(resources_info: Dict):
    if resources_info.get('lambda').get('preheat', False):
        Lambda(resources_info).preheat_function()

############################################
###          ADD EXTRA PROPERTIES        ###
############################################


def _add_extra_aws_properties(scar: Dict, aws_resources: Dict) -> None:
    for resources_info in aws_resources:
        _add_tags(resources_info)
        _add_handler(resources_info)
        _add_account_id(resources_info)
        _add_output(scar)
        _add_config_file_path(scar, resources_info)


def _add_tags(resources_info: Dict):
    resources_info['lambda']['tags'] = {"createdby": "scar", "owner": _get_owner(resources_info)}


def _add_account_id(resources_info: Dict):
    resources_info['iam']['account_id'] = StrUtils.find_expression(resources_info['iam']['role'], _ACCOUNT_ID_REGEX)


def _add_handler(resources_info: Dict):
    resources_info['lambda']['handler'] = f"{resources_info.get('lambda').get('name')}.lambda_handler"


def _add_output(scar_info: Dict):
    scar_info['cli_output'] = response_parser.OutputType.PLAIN_TEXT.value
    if scar_info.get("json", False):
        scar_info['cli_output'] = response_parser.OutputType.JSON.value
    # Override json ouput if both of them are defined
    if scar_info.get("verbose", False):
        scar_info['cli_output'] = response_parser.OutputType.VERBOSE.value
    if scar_info.get("output_file", False):
        scar_info['cli_output'] = response_parser.OutputType.BINARY.value


def _add_config_file_path(scar_info: Dict, resources_info: Dict):
    if scar_info.get("conf_file", False):
        resources_info['lambda']['config_path'] = os.path.dirname(scar_info.get("conf_file"))
        # Update the path of the files based on the path of the yaml (if any)
        if resources_info['lambda'].get('init_script', False):
            resources_info['lambda']['init_script'] = FileUtils.join_paths(resources_info['lambda']['config_path'],
                                                                           resources_info['lambda']['init_script'])
        if resources_info['lambda'].get('image_file', False):
            resources_info['lambda']['image_file'] = FileUtils.join_paths(resources_info['lambda']['config_path'],
                                                                          resources_info['lambda']['image_file'])
        if resources_info['lambda'].get('run_script', False):
            resources_info['lambda']['run_script'] = FileUtils.join_paths(resources_info['lambda']['config_path'],
                                                                          resources_info['lambda']['run_script'])

############################################
###             AWS CONTROLLER           ###
############################################


class AWS(Commands):
    """AWS controller.
    Used to manage all the AWS calls and functionalities."""

    def __init__(self, func_call):
        self.raw_args = FileUtils.load_tmp_config_file()
        AWSValidator.validate_kwargs(self.raw_args)
        self.aws_resources = self.raw_args.get('functions', {}).get('aws', {})
        self.storage_providers = self.raw_args.get('storage_providers', {})
        self.scar_info = self.raw_args.get('scar', {})
        _add_extra_aws_properties(self.scar_info, self.aws_resources)
        # Call the user's command
        getattr(self, func_call)()

############################################
###              AWS COMMANDS            ###
############################################

    @excp.exception(logger)
    def init(self) -> None:
        for resources_info in self.aws_resources:
            resources_info = deepcopy(resources_info)
            _check_function_defined(resources_info)
            # We have to create the gateway before creating the function
            self._create_api_gateway(resources_info)
            # Check the specified supervisor version
            resources_info['lambda']['supervisor']['version'] = SupervisorUtils.check_supervisor_version(
                resources_info.get('lambda').get('supervisor').get('version'))
            self._create_lambda_function(resources_info)
            self._create_log_group(resources_info)
            self._create_s3_buckets(resources_info)
            # The api_gateway permissions must be added after the function is created
            self._add_api_gateway_permissions(resources_info)
            self._create_batch_environment(resources_info)
            _check_preheat_function(resources_info)

    @excp.exception(logger)
    def invoke(self):
        index = 0
        if len(self.aws_resources) > 1:
            index = _choose_function(self.aws_resources)
        if index >= 0:
            resources_info = self.aws_resources[index]
            response = Lambda(resources_info).call_http_endpoint()
            response_parser.parse_http_response(response,
                                                resources_info,
                                                self.scar_info)

    @excp.exception(logger)
    def run(self):
        index = 0
        if len(self.aws_resources) > 1:
            index = _choose_function(self.aws_resources)
        if index >= 0:
            resources_info = self.aws_resources[index]
            using_s3_bucket = False
            if resources_info.get('lambda').get('input', False):
                for storage in resources_info.get('lambda').get('input'):
                    if storage.get('storage_provider') == 's3':
                        print("This function has an associated 'S3' input bucket.")
                        response = input(f"Do you want to launch the function using the files in '{storage.get('path')}'? [Y/n] ")
                        if response in ('Y', 'y'):
                            using_s3_bucket = True
                            self._process_s3_input_bucket_calls(resources_info, storage)
            if not using_s3_bucket:
                response = Lambda(resources_info).launch_lambda_instance()
                if self.scar_info.get("output_file", False):
                    response['OutputFile'] = self.scar_info.get("output_file")
                response['OutputType'] = self.scar_info.get("cli_output")
                response_parser.parse_invocation_response(**response)

    @excp.exception(logger)
    def ls(self):
        # If a bucket is defined, then we list their files
        resources_info = self.aws_resources[0]
        if resources_info.get('lambda').get('input', False):
            file_list = S3(resources_info).get_bucket_file_list()
            for file_info in file_list:
                logger.info(file_info)
        else:
            # Return the resources of the region in the scar's configuration file
            aws_resources = _get_all_functions(self.aws_resources[0])
            response_parser.parse_ls_response(aws_resources, self.scar_info.get('cli_output'))

    @excp.exception(logger)
    def rm(self):
        if self.scar_info.get('all', False):
            # Return the resources of the region in the scar's configuration file
            for resources_info in _get_all_functions(self.aws_resources[0]):
                self._delete_resources(resources_info)
        else:
            index = 0
            if len(self.aws_resources) > 1:
                index = _choose_function(self.aws_resources)
            # -1 means apply to all functions
            if index == -1:
                for resources_info in self.aws_resources:
                    _check_function_not_defined(resources_info)
                    self._delete_resources(resources_info)
            else:
                resources_info = self.aws_resources[index]
                _check_function_not_defined(resources_info)
                self._delete_resources(resources_info)

    @excp.exception(logger)
    def log(self):
        index = 0
        if len(self.aws_resources) > 1:
            index = _choose_function(self.aws_resources)
        # We only return the logs of one function each time
        if index >= 0:
            logger.info(CloudWatchLogs(self.aws_resources[index]).get_aws_logs())

    @excp.exception(logger)
    def put(self):
        self._upload_file_or_folder_to_s3(self.aws_resources[0])

    @excp.exception(logger)
    def get(self):
        self._download_file_or_folder_from_s3(self.aws_resources[0])

#############################################################################
###                   Methods to create AWS resources                     ###
#############################################################################

    @excp.exception(logger)
    def _create_api_gateway(self, resources_info: Dict):
        if resources_info.get("api_gateway", {}).get('name', False):
            APIGateway(resources_info).create_api_gateway()

    @excp.exception(logger)
    def _create_lambda_function(self, resources_info: Dict) -> None:
        lambda_client = Lambda(resources_info)
        response = lambda_client.create_function()
        response_parser.parse_lambda_function_creation_response(response,
                                                                self.scar_info.get('cli_output'),
                                                                lambda_client.get_access_key())

    @excp.exception(logger)
    def _create_log_group(self, resources_info: Dict) -> None:
        cloudwatch_logs = CloudWatchLogs(resources_info)
        response = cloudwatch_logs.create_log_group()
        response_parser.parse_log_group_creation_response(response,
                                                          cloudwatch_logs.get_log_group_name(),
                                                          self.scar_info.get('cli_output'))

    @excp.exception(logger)
    def _create_s3_buckets(self, resources_info: Dict) -> None:
        if resources_info.get('lambda').get('input', False):
            s3_service = S3(resources_info)
            for bucket in resources_info.get('lambda').get('input'):
                if bucket.get('storage_provider') == 's3':
                    bucket_name, folders = s3_service.create_bucket_and_folders(bucket.get('path'))
                    Lambda(resources_info).link_function_and_bucket(bucket_name)
                    s3_service.set_input_bucket_notification(bucket_name, folders)
                    if not folders:
                        logger.info(f'Input bucket "{bucket_name}" successfully created')

        if resources_info.get('lambda').get('output', False):
            s3_service = S3(resources_info)
            for bucket in resources_info.get('lambda').get('output'):
                if bucket.get('storage_provider') == 's3':
                    bucket_name, folders = s3_service.create_bucket_and_folders(bucket.get('path'))
                    if not folders:
                        logger.info(f'Output bucket "{bucket_name}" successfully created')

    @excp.exception(logger)
    def _add_api_gateway_permissions(self, resources_info: Dict):
        if resources_info.get("api_gateway").get('name', False):
            Lambda(resources_info).add_invocation_permission_from_api_gateway()

    @excp.exception(logger)
    def _create_batch_environment(self, resources_info: Dict) -> None:
        mode = resources_info.get('lambda').get('execution_mode')
        if mode in ("batch", "lambda-batch"):
            Batch(resources_info).create_batch_environment()

#############################################################################
###                   Methods to delete AWS resources                     ###
#############################################################################

    def _delete_resources(self, resources_info: Dict) -> None:
        # Delete associated api
        self._delete_api_gateway(resources_info)
        # Delete associated log
        self._delete_logs(resources_info)
        # Delete associated notifications
        self._delete_bucket_notifications(resources_info)
        # Delete function
        self._delete_lambda_function(resources_info)
        # Delete resources batch
        self._delete_batch_resources(resources_info)

    def _delete_api_gateway(self, resources_info: Dict) -> None:
        api_gateway_id = Lambda(resources_info).get_function_configuration().get('Environment').get('Variables').get('API_GATEWAY_ID')
        if api_gateway_id:
            resources_info['lambda']['environment']['Variables']['API_GATEWAY_ID'] = api_gateway_id
            response = APIGateway(resources_info).delete_api_gateway()
            response_parser.parse_delete_api_response(response,
                                                      api_gateway_id,
                                                      self.scar_info.get('cli_output'))

    def _delete_logs(self, resources_info: Dict):
        cloudwatch_logs = CloudWatchLogs(resources_info)
        log_group_name = cloudwatch_logs.get_log_group_name(resources_info.get('lambda').get('name'))
        response = cloudwatch_logs.delete_log_group(log_group_name)
        response_parser.parse_delete_log_response(response,
                                                  log_group_name,
                                                  self.scar_info.get('cli_output'))

    def _delete_bucket_notifications(self, resources_info: Dict) -> None:
        lambda_client = Lambda(resources_info)
        function_name = resources_info.get('lambda').get('name')
        resources_info['lambda']['arn'] = lambda_client.get_function_configuration(function_name).get('FunctionArn')
        resources_info['lambda']['input'] = lambda_client.get_fdl_config(function_name).get('input', False)
        if resources_info.get('lambda').get('input'):
            for input_storage in resources_info.get('lambda').get('input'):
                if input_storage.get('storage_provider') == 's3':
                    bucket_name = input_storage.get('path').split("/", 1)[0]
                    S3(resources_info).delete_bucket_notification(bucket_name)

    def _delete_lambda_function(self, resources_info: Dict) -> None:
        response = Lambda(resources_info).delete_function()
        response_parser.parse_delete_function_response(response,
                                                       resources_info.get('lambda').get('name'),
                                                       self.scar_info.get('cli_output'))

    def _delete_batch_resources(self, resources_info: Dict) -> None:
        batch = Batch(resources_info)
        if batch.exist_compute_environments():
            batch.delete_compute_environment()

###########################################################
###            Methods to manage S3 resources           ###
###########################################################

    def _process_s3_input_bucket_calls(self, resources_info: Dict, storage: Dict) -> None:
        s3_service = S3(resources_info)
        lambda_service = Lambda(resources_info)
        s3_file_list = s3_service.get_bucket_file_list(storage)
        bucket_name, _ = get_bucket_and_folders(storage.get('path'))
        logger.info(f"Files found: '{s3_file_list}'")
        # First do a request response invocation to prepare the lambda environment
        if s3_file_list:
            s3_event = s3_service.get_s3_event(bucket_name, s3_file_list.pop(0))
            lambda_service.launch_request_response_event(s3_event)
        # If the list has more elements, invoke functions asynchronously
        if s3_file_list:
            s3_event_list = s3_service.get_s3_event_list(bucket_name, s3_file_list)
            lambda_service.process_asynchronous_lambda_invocations(s3_event_list)

    def _upload_file_or_folder_to_s3(self, resources_info: Dict) -> None:
        path_to_upload = self.scar_info.get('path')
        files = [path_to_upload]
        if os.path.isdir(path_to_upload):
            files = FileUtils.get_all_files_in_directory(path_to_upload)
        s3_service = S3(resources_info)
        storage_path = resources_info.get('lambda').get('input')[0].get('path')
        bucket, folder = s3_service.create_bucket_and_folders(storage_path)
        for file_path in files:
            s3_service.upload_file(bucket=bucket, folder_name=folder, file_path=file_path)

    def _get_download_file_path(self, file_key=None):
        file_path = file_key
        if self.scar_info.get('path', False):
            file_path = FileUtils.join_paths(self.scar_info.get('path'), file_path)
        return file_path

    def _download_file_or_folder_from_s3(self, resources_info: Dict) -> None:

        s3_service = S3(resources_info)
        s3_file_list = s3_service.get_bucket_file_list()
        for s3_file in s3_file_list:
            # Avoid download s3 'folders'
            if not s3_file.endswith('/'):
                file_path = self._get_download_file_path(file_key=s3_file)
                # make sure the path folders are created
                dir_path = os.path.dirname(file_path)
                if dir_path and not os.path.isdir(dir_path):
                    os.makedirs(dir_path, exist_ok=True)
                bucket, _ = get_bucket_and_folders(resources_info.get('lambda').get('input')[0].get('path'))
                s3_service.download_file(bucket, s3_file, file_path)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with classes and methods used for parameter validation."""

from scar.exceptions import ValidatorError, S3CodeSizeError, \
                            FunctionCodeSizeError, InvocationPayloadError
from scar.utils import FileUtils, StrUtils

VALID_LAMBDA_NAME_REGEX = (r"(arn:(aws[a-zA-Z-]*)?:lambda:)?([a-z]{2}(-gov)?-[a-z]+-\d{1}:)?("
                           r"\d{12}:)?(function:)?([a-zA-Z0-9-_]+)(:(\$LATEST|[a-zA-Z0-9-_]+))?")
KB = 1024
MB = KB * KB
MAX_POST_BODY_SIZE = MB * 6
MAX_POST_BODY_SIZE_ASYNC = KB * 95


class AWSValidator():
    """Class with methods to validate AWS properties."""

    @staticmethod
    def validate_kwargs(cls, **kwargs):
        aws_functions = kwargs.get('functions', {}).get('aws', {})
        for function in aws_functions:
            if 'iam' in function:
                cls.validate_iam(function['iam'])
            if 'lambda' in function:
                cls.validate_lambda(function['lambda'])
            if 'batch' in function:
                cls.validate_batch(function['batch'])

    @staticmethod
    def validate_iam(iam_properties):
        if ("role" not in iam_properties) or (iam_properties["role"] == ""):
            error_msg = ("Please, specify a valid iam role in the "
                         "configuration file (usually located in ~/.scar/scar.cfg).")
            raise ValidatorError(parameter='iam_role',
                                 parameter_value=iam_properties,
                                 error_msg=error_msg)

    @staticmethod
    def validate_lambda(cls, lambda_properties):
        if 'name' in lambda_properties:
            cls.validate_function_name(lambda_properties['name'])
        if 'memory' in lambda_properties:
            cls.validate_memory(lambda_properties['memory'])
        if 'time' in lambda_properties:
            cls.validate_time(lambda_properties['time'])

    @staticmethod
    def validate_batch(cls, batch_properties):
        if 'vcpus' in batch_properties:
            cls.validate_batch_vcpus(batch_properties['vcpus'])
        if 'memory' in batch_properties:
            cls.validate_batch_memory(batch_properties['memory'])
        if 'compute_resources' in batch_properties and \
            'comp_type' in batch_properties['compute_resources']:
            cls.validate_batch_comp_type(batch_properties['compute_resources']['comp_type'])

    @staticmethod
    def validate_time(lambda_time):
        if (lambda_time <= 0) or (lambda_time > 900):
            error_msg = 'Please, set a value between 0 and 900.'
            raise ValidatorError(parameter='lambda_time',
                                 parameter_value=lambda_time,
                                 error_msg=error_msg)

    @staticmethod
    def validate_memory(lambda_memory):
        if (lambda_memory < 128) or (lambda_memory > 3008):
            error_msg = 'Please, set a value between 128 and 3008.'
            raise ValidatorError(parameter='lambda_memory',
                                 parameter_value=lambda_memory,
                                 error_msg=error_msg)

    @staticmethod
    def validate_function_name(function_name):
        if not StrUtils.find_expression(function_name, VALID_LAMBDA_NAME_REGEX):
            error_msg = ("Find name restrictions in: https://docs.aws.amazon.com/lambda/latest/"
                         "dg/API_CreateFunction.html#SSS-CreateFunction-request-FunctionName")
            raise ValidatorError(parameter='function_name',
                                 parameter_value=function_name,
                                 error_msg=error_msg)

    @staticmethod
    def validate_function_code_size(scar_folder, max_payload_size):
        if FileUtils.get_tree_size(scar_folder) > max_payload_size:
            raise FunctionCodeSizeError(code_size='50MB')

    @staticmethod
    def validate_s3_code_size(scar_folder, max_s3_payload_size):
        if FileUtils.get_tree_size(scar_folder) > max_s3_payload_size:
            raise S3CodeSizeError(code_size='250MB')

    @staticmethod
    def validate_http_payload_size(file_path, async_call=False):
        file_size = FileUtils.get_file_size(file_path)
        if file_size > MAX_POST_BODY_SIZE:
            filesize = '{0:.2f}MB'.format(file_size / MB)
            maxsize = '{0:.2f}MB'.format(MAX_POST_BODY_SIZE / MB)
            raise InvocationPayloadError(file_size=filesize, max_size=maxsize)
        if async_call and file_size > MAX_POST_BODY_SIZE_ASYNC:
            filesize = '{0:.2f}KB'.format(file_size / KB)
            maxsize = '{0:.2f}KB'.format(MAX_POST_BODY_SIZE_ASYNC / KB)
            raise InvocationPayloadError(file_size=filesize, max_size=maxsize)

    @staticmethod
    def validate_batch_vcpus(batch_vcpus):
        if batch_vcpus < 1:
            error_msg = 'Please, set at least one vCPU.'
            raise ValidatorError(parameter='batch_vcpus',
                                 parameter_value=batch_vcpus,
                                 error_msg=error_msg)

    @staticmethod
    def validate_batch_memory(batch_memory):
        if batch_memory < 4:
            error_msg = 'Please, set a value greater than 4.'
            raise ValidatorError(parameter='batch_memory',
                                 parameter_value=batch_memory,
                                 error_msg=error_msg)

    @staticmethod
    def validate_batch_comp_type(batch_comp_type):
        if batch_comp_type not in ['SPOT', 'EC2']:
            error_msg = 'Please, set a valid compute environment type ("EC2" or "SPOT")'
            raise ValidatorError(parameter='batch_comp_type',
                                 parameter_value=batch_comp_type,
                                 error_msg=error_msg)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with methods shared by all the classes."""

import base64
import json
import os
import re
import shutil
import subprocess
import tarfile
import tempfile
import uuid
import sys
from copy import deepcopy
from zipfile import ZipFile
from io import BytesIO
from typing import Optional, Dict, List, Generator, Union, Any
from distutils import dir_util
from packaging import version
import yaml
import scar.logger as logger
import scar.http.request as request
from scar.exceptions import GitHubTagNotFoundError, YamlFileNotFoundError

COMMANDS = ['scar-config']


def lazy_property(func):
    # Skipped type hinting: https://github.com/python/mypy/issues/3157
    """ A decorator that makes a property lazy-evaluated."""
    attr_name = '_lazy_' + func.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, func(self))
        return getattr(self, attr_name)

    return _lazy_property


class SysUtils:
    """Common methods for system management."""

    @staticmethod
    def is_variable_in_environment(variable: str) -> bool:
        """Checks if a variable is in the system environment."""
        return variable in os.environ

    @staticmethod
    def set_environment_variable(key: str, variable: Any) -> None:
        """Sets a system environment variable."""
        if key and variable:
            os.environ[key] = variable

    @staticmethod
    def get_environment_variable(variable: str) -> str:
        """Returns the value of system environment variable
        or an empty string if not found."""
        return os.environ.get(variable, '')

    @staticmethod
    def delete_environment_variable(variable: str) -> None:
        """Delete a system environment variable."""
        if SysUtils.is_variable_in_environment(variable):
            del os.environ[variable]

    @staticmethod
    def execute_command_with_msg(command: List[str], cmd_wd: Optional[str]=None,
                                 cli_msg: str='') -> str:
        """Execute the specified command and return the result."""
        cmd_out = subprocess.check_output(command, cwd=cmd_wd).decode('utf-8')
        logger.debug(cmd_out)
        logger.info(cli_msg)
        return cmd_out[:-1]

    @staticmethod
    def get_user_home_path() -> str:
        """Returns the path of the current user's home."""
        return os.path.expanduser("~")

    @staticmethod
    def finish_scar_execution() -> None:
        """Finishes the program execution."""
        logger.end_execution_trace()
        sys.exit(0)


class DataTypesUtils:
    """Common methods for data types management."""

    @staticmethod
    def merge_dicts(dict1: Dict, dict2: Dict) -> Dict:
        """Merge 'dict1' and 'dict2' dicts into 'dict1'.
        'dict2' has precedence over 'dict1'."""
        for key, val in dict2.items():
            if val is not None:
                if isinstance(val, dict) and key in dict1:
                    dict1[key] = DataTypesUtils.merge_dicts(dict1[key], val)
                elif isinstance(val, list) and key in dict1:
                    dict1[key] += val
                else:
                    dict1[key] = val
        return dict1

    @staticmethod
    def merge_dicts_with_copy(dict1: Dict, dict2: Dict) -> Dict:
        """Merge 'dict1' and 'dict2' dicts into a new Dict.
        'dict2' has precedence over 'dict1'."""
        result = deepcopy(dict1)
        for key, val in dict2.items():
            if val is not None:
                if isinstance(val, dict) and key in result:
                    result[key] = DataTypesUtils.merge_dicts_with_copy(result[key], val)
                elif isinstance(val, list) and key in result:
                    result[key] += val
                else:
                    result[key] = val
        return result

    @staticmethod
    def divide_list_in_chunks(elements: List, chunk_size: int) -> Generator[List, None, None]:
        """Yield successive n-sized chunks from th elements list."""
        if not elements:
            yield []
        for i in range(0, len(elements), chunk_size):
            yield elements[i:i + chunk_size]

    @staticmethod
    def parse_arg_list(arg_keys: List, cmd_args: Dict) -> Dict:
        """Parse an argument dictionary filtering by the names
        specified in a list."""
        result = {}
        for key in arg_keys:
            if isinstance(key, tuple):
                if key[0] in cmd_args and cmd_args[key[0]]:
                    result[key[1]] = cmd_args[key[0]]
            else:
                if key in cmd_args and cmd_args[key]:
                    result[key] = cmd_args[key]
        return result


class FileUtils:
    """Common methods for file and directory management."""

    @staticmethod
    def copy_file(source: str, dest: str) -> None:
        """Copy file to specified destination."""
        shutil.copy(source, dest)

    @staticmethod
    def copy_dir(source: str, dest: str) -> None:
        """Copy directory to specified destination."""
        dir_util.copy_tree(source, dest)

    @staticmethod
    def create_folder(folder_name):
        """Creates a system folder.
        Does nothing if the folder exists."""
        os.makedirs(folder_name, exist_ok=True)

    @staticmethod
    def get_scar_root_path() -> str:
        """Returns the root path of the project."""
        return os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    @staticmethod
    def join_paths(*paths: str) -> str:
        """Returns the strings passed joined as a system path."""
        return os.path.join(*paths)

    @staticmethod
    def get_tmp_dir() -> str:
        """Gets the directory where the temporal
        folder of the system is located."""
        return tempfile.gettempdir()

    @staticmethod
    def create_tmp_dir() -> tempfile.TemporaryDirectory:
        """Creates a directory in the temporal folder of the system.
        When the context is finished, the folder is automatically deleted."""
        return tempfile.TemporaryDirectory()

    @staticmethod
    def create_tmp_file(**kwargs) -> tempfile.NamedTemporaryFile:
        """Creates a directory in the temporal folder of the system.
        When the context is finished, the folder is automatically deleted."""
        return tempfile.NamedTemporaryFile(**kwargs)

    @staticmethod
    def get_tree_size(path: str) -> int:
        """Return total size of files in given path and subdirs."""
        total = 0
        for entry in os.scandir(path):
            if entry.is_dir(follow_symlinks=False):
                total += FileUtils.get_tree_size(entry.path)
            else:
                total += entry.stat(follow_symlinks=False).st_size
        return total

    @staticmethod
    def get_all_files_in_directory(dir_path: str) -> List[str]:
        """Returns a list with all the file paths in
        the specified directory and subdirectories."""
        files = []
        for dirname, _, filenames in os.walk(dir_path):
            for filename in filenames:
                files.append(os.path.join(dirname, filename))
        return files

    @staticmethod
    def get_file_size(file_path: str) -> int:
        """Returns the file size in bytes"""
        return os.stat(file_path).st_size

    @staticmethod
    def create_file_with_content(path: str,
                                 content: Optional[Union[str, bytes]],
                                 mode: str='w') -> None:
        """Creates a new file with the passed content.
        If the content is a dictionary, first is converted to a string."""
        with open(path, mode) as fwc:
            if isinstance(content, dict):
                content = json.dumps(content)
            fwc.write(content)

    @staticmethod
    def read_file(file_path: str, mode: str='r') -> Optional[Union[str, bytes]]:
        """Reads the whole specified file and returns the content."""
        with open(file_path, mode) as content_file:
            return content_file.read()

    @staticmethod
    def delete_file(path: str) -> None:
        """Delete the specified file."""
        if os.path.isfile(path):
            os.remove(path)

    @staticmethod
    def delete_folder(path: str) -> None:
        """Delete a folder with all its contents."""
        shutil.rmtree(path)

    @staticmethod
    def create_tar_gz(files_to_archive: List[str], destination_tar_path: str) -> str:
        """Create a .tar.gz file with the passed list of files."""
        with tarfile.open(destination_tar_path, 'w:gz') as tar:
            for file_path in files_to_archive:
                tar.add(file_path, arcname=os.path.basename(file_path))
        return destination_tar_path

    @staticmethod
    def extract_tar_gz(tar_path: str, destination_path: str) -> None:
        """Extract the content of a .tar.gz file in the specified path."""
        with tarfile.open(tar_path, 'r:gz') as tar:
            tar.extractall(path=destination_path)

    @staticmethod
    def unzip_folder(zip_path: str, folder_where_unzip_path: str, msg: str='') -> None:
        """Must use the unzip binary to preserve the file properties and the symlinks."""
        zip_exe = '/usr/bin/unzip'
        SysUtils.execute_command_with_msg([zip_exe, zip_path],
                                          cmd_wd=folder_where_unzip_path,
                                          cli_msg=msg)

    @staticmethod
    def zip_folder(zip_path: str, folder_to_zip_path: str, msg: str='') -> None:
        """Must use the zip binary to preserve the file properties and the symlinks."""
        zip_exe = '/usr/bin/zip'
        SysUtils.execute_command_with_msg([zip_exe, '-r9y', zip_path, '.'],
                                          cmd_wd=folder_to_zip_path,
                                          cli_msg=msg)

    @staticmethod
    def is_file(file_path: str):
        """Test whether a path is a regular file."""
        return os.path.isfile(file_path)

    @staticmethod
    def load_yaml(file_path: str) -> Dict:
        """Returns the content of a YAML file as a Dict."""
        if os.path.isfile(file_path):
            with open(file_path) as cfg_file:
                return yaml.safe_load(cfg_file)
        else:
            raise YamlFileNotFoundError(file_path=file_path)

    @staticmethod
    def write_yaml(file_path: str, content: Dict) -> None:
        with open(file_path, 'w') as cfg_file:
            yaml.safe_dump(content, cfg_file)

    @staticmethod
    def create_tmp_config_file(cfg_args, ConfigFileParser):
        cfg_path = ConfigFileParser.tmp_yaml_file_path
        os.environ['SCAR_TMP_CFG'] = cfg_path
        FileUtils.write_yaml(cfg_path, cfg_args)

    @staticmethod
    def load_tmp_config_file():
        return FileUtils.load_yaml(os.environ['SCAR_TMP_CFG'])

    @staticmethod
    def get_file_name(file_path: str) -> str:
        return os.path.basename(file_path)

    @staticmethod
    def extract_zip_from_url(url: str, dest_path: str) -> None:
        with ZipFile(BytesIO(url)) as thezip:
            thezip.extractall(dest_path)


class StrUtils:
    """Common methods for string management."""

    @staticmethod
    def decode_base64(value: Union[bytes, str]) -> bytes:
        """Decode a Base64 encoded bytes-like object or
        ASCII string and return the decoded bytes"""
        return base64.b64decode(value)

    @staticmethod
    def encode_base64(value: bytes) -> bytes:
        """Encode a bytes-like object using Base64
        and return the encoded bytes."""
        return base64.b64encode(value)

    @staticmethod
    def base64_to_utf8_string(value: Union[bytes, str]) -> str:
        """Decode a Base64 encoded bytes-like object or ASCII
        string and return the decoded value as a string."""
        return StrUtils.decode_base64(value).decode('utf-8')

    @staticmethod
    def utf8_to_base64_string(value: str) -> str:
        """Encode a 'utf-8' string using Base64 and return
        the encoded value as a string."""
        return StrUtils.encode_base64(bytes(value, 'utf-8')).decode('utf-8')

    @staticmethod
    def bytes_to_base64str(value, encoding='utf-8') -> str:
        """Encode a 'utf-8' string using Base64 and return
        the encoded value as a string."""
        return StrUtils.encode_base64(value).decode(encoding)

    @staticmethod
    def dict_to_base64_string(value: Dict) -> str:
        """Encodes a dictionary to base64 and returns a string."""
        return StrUtils.utf8_to_base64_string(json.dumps(value))

    @staticmethod
    def find_expression(string_to_search: str, rgx_pattern: str) -> Optional[str]:
        """Returns the first group that matches the rgx_pattern in the string_to_search."""
        if string_to_search:
            pattern = re.compile(rgx_pattern)
            match = pattern.search(string_to_search)
            if match:
                return match.group()
        return None

    @staticmethod
    def get_random_uuid4_str() -> str:
        """Returns a random generated uuid4 string."""
        return str(uuid.uuid4())

    @staticmethod
    def compare_versions(ver1: str, ver2: str) -> int:
        """Returns value < 0 to indicate that ver1 is less than ver2.
        Returns value > 0 to indicate that ver1 is greater than ver2.
        Returns value == 0 to indicate that ver1 is equal to ver2."""
        res = 0
        if version.parse(ver1) < version.parse(ver2):
            res = -1
        elif version.parse(ver1) > version.parse(ver2):
            res = 1
        return res


class GitHubUtils:
    """Common methods for GitHub API Queries.
    https://developer.github.com/v3/repos/releases/"""

    @staticmethod
    def get_latest_release(user: str, project: str) -> str:
        """Get the tag of the latest release in a repository."""
        url = f'https://api.github.com/repos/{user}/{project}/releases/latest'
        response = json.loads(request.get_file(url))
        return response.get('tag_name', '')

    @staticmethod
    def exists_release_in_repo(user: str, project: str, tag_name: str) -> bool:
        """Check if a tagged release exists in a repository."""
        url = f'https://api.github.com/repos/{user}/{project}/releases/tags/{tag_name}'
        response = request.get_file(url)
        if not response:
            return False
        response = json.loads(response)
        if 'message' in response and response['message'] == 'Not Found':
            return False
        return True

    @staticmethod
    def get_asset_url(user: str, project: str, asset_name: str,
                      tag_name: str='latest') -> Optional[str]:
        """Get the download asset url from the specified github tagged project."""
        if tag_name == 'latest':
            url = f'https://api.github.com/repos/{user}/{project}/releases/latest'
        else:
            if GitHubUtils.exists_release_in_repo(user, project, tag_name):
                url = f'https://api.github.com/repos/{user}/{project}/releases/tags/{tag_name}'
            else:
                raise GitHubTagNotFoundError(tag=tag_name)
        response = json.loads(request.get_file(url))
        if isinstance(response, dict) and 'assets' in response:
            for asset in response['assets']:
                if asset['name'] == asset_name:
                    return asset['browser_download_url']
        return None

    @staticmethod
    def get_source_code_url(user: str, project: str, tag_name: str='latest') -> str:
        """Get the source code's url from the specified github tagged project."""
        source_url = ""
        repo_url = ""
        if tag_name == 'latest':
            repo_url = f'https://api.github.com/repos/{user}/{project}/releases/latest'
        else:
            if GitHubUtils.exists_release_in_repo(user, project, tag_name):
                repo_url = f'https://api.github.com/repos/{user}/{project}/releases/tags/{tag_name}'
            else:
                raise GitHubTagNotFoundError(tag=tag_name)
        if repo_url:
            response = json.loads(request.get_file(repo_url))
            if isinstance(response, dict):
                source_url = response.get('zipball_url')
        return source_url


class SupervisorUtils:
    """Common methods for FaaS Supervisor management.
    https://github.com/grycap/faas-supervisor/"""

    _SUPERVISOR_GITHUB_REPO = 'faas-supervisor'
    _SUPERVISOR_GITHUB_USER = 'grycap'
    _SUPERVISOR_GITHUB_ASSET_NAME = 'supervisor'

    @classmethod
    def download_supervisor(cls, supervisor_version: str, path: str) -> str:
        """Downloads the FaaS Supervisor .zip package to the specified path."""
        supervisor_zip_path = FileUtils.join_paths(path, 'faas-supervisor.zip')
        supervisor_zip_url = GitHubUtils.get_source_code_url(
            cls._SUPERVISOR_GITHUB_USER,
            cls._SUPERVISOR_GITHUB_REPO,
            supervisor_version)
        with open(supervisor_zip_path, "wb") as thezip:
            thezip.write(request.get_file(supervisor_zip_url))
        return supervisor_zip_path

    @classmethod
    def check_supervisor_version(cls, supervisor_version: str) -> str:
        """Checks if the specified version exists in FaaS Supervisor's GitHub
        repository. Returns the version if exists and 'latest' if not."""
        if GitHubUtils.exists_release_in_repo(cls._SUPERVISOR_GITHUB_USER,
                                              cls._SUPERVISOR_GITHUB_REPO,
                                              supervisor_version):
            return supervisor_version
        latest_version = SupervisorUtils.get_latest_release()
        if supervisor_version != 'latest':
            logger.info('Defined supervisor version does not exists.')
        logger.info(f'Using latest supervisor release: \'{latest_version}\'.')
        return latest_version

    @classmethod
    def get_supervisor_binary_url(cls, supervisor_version: str) -> str:
        """Returns the supervisor's binary download url."""
        return GitHubUtils.get_asset_url(cls._SUPERVISOR_GITHUB_USER,
                                         cls._SUPERVISOR_GITHUB_REPO,
                                         cls._SUPERVISOR_GITHUB_ASSET_NAME,
                                         supervisor_version)

    @classmethod
    def get_latest_release(cls) -> str:
        """Returns the latest FaaS Supervisor version."""
        return GitHubUtils.get_latest_release(cls._SUPERVISOR_GITHUB_USER,
                                              cls._SUPERVISOR_GITHUB_REPO)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with the class necessary to manage the
resource groups listing."""

from typing import List
from scar.providers.aws.clients import BotoClient
from scar.exceptions import exception
import scar.logger as logger


class ResourceGroupsClient(BotoClient):
    """A low-level client representing aws Resource Groups Tagging API.
    DOC_URL: https://boto3.readthedocs.io/en/latest/reference/
             services/resourcegroupstaggingapi.html"""

    # Parameter used by the parent to create the appropriate boto3 client
    _BOTO_CLIENT_NAME = 'resourcegroupstaggingapi'

    @exception(logger)
    def get_tagged_resources(self, tag_filters: List,
                             resource_type_filters: List,
                             next_token: str = None) -> List:
        """Returns all the tagged resources that are associated with
        the specified tags (keys and values) located in the specified
        region for the AWS account."""
        resource_list = []
        kwargs = {"TagFilters" : tag_filters,
                  "ResourceTypeFilters" : resource_type_filters}
        if next_token:
            kwargs['PaginationToken'] = next_token
        response = self.client.get_resources(**kwargs)
        if 'ResourceTagMappingList' in response:
            resource_list.extend(response['ResourceTagMappingList'])
        # Retrieve all the remaining resources recursively
        if 'PaginationToken' in response and response['PaginationToken']:
            resource_list.extend(self.get_tagged_resources(tag_filters,
                                                           resource_type_filters,
                                                           kwargs['PaginationToken']))
        return resource_list



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from _ast import Or
"""Module with methods and classes to create the function deployment package."""

from typing import Dict
from zipfile import ZipFile
from scar.providers.aws.udocker import Udocker
from scar.providers.aws.validators import AWSValidator
from scar.exceptions import exception
import scar.logger as logger
from scar.utils import FileUtils


def create_function_config(resources_info):
    function_cfg = {'storage_providers': FileUtils.load_tmp_config_file().get('storage_providers', {})}
    function_cfg.update(resources_info.get('lambda'))
    return function_cfg


class FunctionPackager():
    """Class to manage the deployment package creation."""

    def __init__(self, resources_info: Dict, supervisor_zip_path: str):
        self.resources_info = resources_info
        self.supervisor_zip_path = supervisor_zip_path
        # Temporal folder to store the supervisor and udocker files
        self.tmp_payload_folder = FileUtils.create_tmp_dir()

    @exception(logger)
    def create_zip(self, lambda_payload_path: str) -> None:
        """Creates the lambda function deployment package."""
        self._extract_handler_code()
        self._manage_udocker_images()
        self._add_init_script()
        self._add_extra_payload()
        self._copy_function_configuration()
        self._zip_scar_folder(lambda_payload_path)
        self._check_code_size()

    def _extract_handler_code(self) -> None:
        function_handler_dest = FileUtils.join_paths(self.tmp_payload_folder.name, f"{self.resources_info.get('lambda').get('name')}.py")
        file_path = ""
        with ZipFile(self.supervisor_zip_path) as thezip:
            for file in thezip.namelist():
                if file.endswith("function_handler.py"):
                    file_path = FileUtils.join_paths(FileUtils.get_tmp_dir(), file)
                    # Extracts the complete folder structure and the file (cannot avoid)
                    thezip.extract(file, FileUtils.get_tmp_dir())
                    break
        if file_path:
            # Copy only the handler to the payload folder
            FileUtils.copy_file(file_path, function_handler_dest)

    def _copy_function_configuration(self):
        cfg_file_path = FileUtils.join_paths(self.tmp_payload_folder.name, "function_config.yaml")
        function_cfg = create_function_config(self.resources_info)
        FileUtils.write_yaml(cfg_file_path, function_cfg)

    def _manage_udocker_images(self):
        if self.resources_info.get('lambda').get('container').get('image_file', False) or \
           self.resources_info.get('lambda').get('deployment').get('bucket', False):
            Udocker(self.resources_info, self.tmp_payload_folder.name, self.supervisor_zip_path).prepare_udocker_image()

    def _add_init_script(self) -> None:
        """Copy the init script defined by the user to the payload folder."""
        if self.resources_info.get('lambda').get('init_script', False):
            init_script_path = self.resources_info.get('lambda').get('init_script')
            FileUtils.copy_file(init_script_path,
                                FileUtils.join_paths(self.tmp_payload_folder.name,
                                                     FileUtils.get_file_name(init_script_path)))

    def _add_extra_payload(self) -> None:
        if self.resources_info.get('lambda').get('extra_payload', False):
            payload_path = self.resources_info.get('lambda').get('extra_payload')
            logger.info(f"Adding extra payload '{payload_path}'")
            if FileUtils.is_file(payload_path):
                FileUtils.copy_file(self.resources_info.get('lambda').get('extra_payload'),
                                    self.tmp_payload_folder.name)
            else:
                FileUtils.copy_dir(self.resources_info.get('lambda').get('extra_payload'),
                                   self.tmp_payload_folder.name)
            del(self.resources_info['lambda']['extra_payload'])

    def _zip_scar_folder(self, lambda_payload_path: str) -> None:
        """Zips the tmp folder with all the function's files and
        save it in the expected path of the payload."""
        FileUtils.zip_folder(lambda_payload_path,
                             self.tmp_payload_folder.name,
                             "Creating function package.")

    def _check_code_size(self):
        # Check if the code size fits within the AWS limits
        if self.resources_info.get('lambda').get('deployment').get('bucket', False):
            AWSValidator.validate_s3_code_size(self.tmp_payload_folder.name,
                                               self.resources_info.get('lambda').get('deployment').get('max_s3_payload_size'))
        else:
            AWSValidator.validate_function_code_size(self.tmp_payload_folder.name,
                                                     self.resources_info.get('lambda').get('deployment').get('max_payload_size'))



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with methods and classes to manage the Lambda layers."""
import shutil
from typing import Dict, List
import zipfile
import scar.logger as logger
from scar.utils import FileUtils
from scar.providers.aws.clients.lambdafunction import LambdaClient


class Layer():
    """Class used for layer management."""

    def __init__(self, lambda_client: LambdaClient) -> None:
        self.lambda_client = lambda_client

    def _find(self, layer_name: str) -> Dict:
        """Returns the layer information that matches the layer name passed."""
        for layer in self.lambda_client.list_layers():
            if layer.get('LayerName', '') == layer_name:
                return layer
        return {}

    def create(self, **kwargs: Dict) -> Dict:
        """Creates a new layer with the arguments passed."""
        return self.lambda_client.publish_layer_version(**kwargs)

    def exists(self, layer_name: str) -> bool:
        """Checks layer name for existence."""
        if self._find(layer_name):
            return True
        return False

    def list_versions(self, layer_name: str) -> List:
        return self.lambda_client.list_layer_versions(layer_name)

    def delete(self, **kwargs: Dict) -> Dict:
        """Deletes a layer."""
        layer_args = {'LayerName': kwargs['name']}
        if 'version' in kwargs:
            layer_args['VersionNumber'] = int(kwargs['version'])
        else:
            version_info = self.get_latest_layer_info(kwargs['name'])
            layer_args['VersionNumber'] = version_info.get('Version', -1)
        return self.lambda_client.delete_layer_version(**layer_args)

    def get_latest_layer_info(self, layer_name: str) -> Dict:
        """Returns the latest matching version of the layer with 'layer_name'."""
        layer = self._find(layer_name)
        return layer['LatestMatchingVersion'] if layer else {}


class LambdaLayers():
    """"Class used to manage the lambda supervisor layer."""

    # To avoid circular inheritance we need to receive the LambdaClient
    def __init__(self, resources_info: Dict, lambda_client: LambdaClient, supervisor_zip_path: str):
        self.resources_info = resources_info
        self.supervisor_zip_path = supervisor_zip_path
        self.layer_name = resources_info.get('lambda').get('supervisor').get('layer_name')
        self.supervisor_version = resources_info.get('lambda').get('supervisor').get('version')
        self.layer = Layer(lambda_client)

    def _get_supervisor_layer_props(self, layer_zip_path: str) -> Dict:
        return {'LayerName': self.layer_name,
                'Description': self.supervisor_version,
                'Content': {'ZipFile': FileUtils.read_file(layer_zip_path, mode="rb")},
                'CompatibleRuntimes': ['python3.8', 'python3.7'],
                'LicenseInfo': self.resources_info.get('lambda').get('supervisor').get('license_info')}

    def _create_layer(self) -> str:
        # Create tmp folders
        tmp_path = FileUtils.create_tmp_dir()
        layer_code_path = FileUtils.create_tmp_dir()
        # Extract 'extra' and 'faassupervisor' from supervisor_zip_path
        with zipfile.ZipFile(self.supervisor_zip_path) as thezip:
            for file in thezip.namelist():
                # Remove the parent folder path
                parent_folder, file_name = file.split('/', 1)
                if file_name.startswith('extra') or file_name.startswith('faassupervisor'):
                    thezip.extract(file, tmp_path.name)
        # Extract content of 'extra' files in layer_code_path
        extra_folder_path = FileUtils.join_paths(tmp_path.name, parent_folder, 'extra')
        files = FileUtils.get_all_files_in_directory(extra_folder_path)
        for file_path in files:
            FileUtils.unzip_folder(file_path, layer_code_path.name)
        # Copy 'faassupervisor' to layer_code_path
        supervisor_folder_path = FileUtils.join_paths(tmp_path.name, parent_folder, 'faassupervisor')
        shutil.move(supervisor_folder_path, FileUtils.join_paths(layer_code_path.name, 'python', 'faassupervisor'))
        # Create layer zip with content of layer_code_path
        layer_zip_path = FileUtils.join_paths(tmp_path.name, f'{self.layer_name}.zip')
        FileUtils.zip_folder(layer_zip_path, layer_code_path.name)
        # Register the layer
        props = self._get_supervisor_layer_props(layer_zip_path)
        response = self.layer.create(**props)
        return response['LayerVersionArn']

    def _is_supervisor_created(self) -> bool:
        return self.layer.exists(self.layer_name)

    def _is_supervisor_version_created(self) -> str:
        versions = self.layer.list_versions(self.layer_name)
        for version in versions:
            if 'Description' in version:
                if version['Description'] == self.supervisor_version:
                    return version['LayerVersionArn']
        return ''

    def get_supervisor_layer_arn(self) -> str:
        """Returns the ARN of the specified supervisor layer version.
        If the layer or version doesn't exists, creates the layer."""
        if self._is_supervisor_created():
            is_created = self._is_supervisor_version_created()
            if is_created != '':
                logger.info(f'Using existent \'{self.layer_name}\' layer.')
                return is_created
        logger.info((f'Creating lambda layer with \'{self.layer_name}\''
                     f' version \'{self.supervisor_version}\'.'))
        return self._create_layer()



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with methods and classes to manage AWS Resource Groups"""

from typing import List
from botocore.exceptions import ClientError
from scar.providers.aws import GenericClient
import scar.logger as logger


class ResourceGroups(GenericClient):
    """Class to manage AWS Resource Groups"""

    def __init__(self, resources_info) -> None:
        super().__init__(resources_info.get('lambda'))

    def get_resource_arn_list(self, iam_user_id: str, resource_type: str = 'lambda') -> List:
        """Returns a list of ARNs filtered by the resource_type
        passed and the tags created by scar."""
        try:
            # Creation of a function_info filter by tags
            tag_filters = [{'Key': 'owner', 'Values': [iam_user_id]},
                           {'Key': 'createdby', 'Values': ['scar']}]
            resource_type_filters = [resource_type]
            tagged_resources = self.client.get_tagged_resources(tag_filters, resource_type_filters)
            return [function_info['ResourceARN'] for function_info in tagged_resources]
        except ClientError as cerr:
            logger.error("Error getting function_info arn by tag",
                         f"Error getting function_info arn by tag: {cerr}")
            raise cerr



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from scar.parser.fdl import FDLParser


class Test(unittest.TestCase):

    def testFDL(self):
        ''' Expected return value (except random ids):
        [{'env_vars':{
            'STORAGE_AUTH_MINIO_PASS_TMPXCMCNA9S': 'mpass',
            'STORAGE_AUTH_MINIO_USER_TMPXCMCNA9S': 'muser',
            'STORAGE_PATH_INPUT_TMPOTOWSDYE': 's3-bucket/test1',
            'STORAGE_PATH_INPUT_TMPXCMCNA9S': 'my-bucket/test',
            'STORAGE_PATH_OUTPUT_TMPOTOWSDYE': 's3-bucket/test1-output',
            'STORAGE_PATH_OUTPUT_TMPXCMCNA9S': 'my-bucket/test-output',
            'STORAGE_PATH_SUFIX_TMPOTOWSDYE': 'avi',
            'STORAGE_PATH_SUFIX_TMPXCMCNA9S': 'wav:srt'},
          'name': 'function1'},
         {'env_vars': {
            'STORAGE_AUTH_MINIO_PASS_TMPXCMCNA9S': 'mpass',
            'STORAGE_AUTH_MINIO_USER_TMPXCMCNA9S': 'muser',
            'STORAGE_PATH_INPUT_TMPXCMCNA9S': 'my-bucket2/test',
            'STORAGE_PATH_OUTPUT_TMPXCMCNA9S': 'my-bucket2/test-output',
            'STORAGE_PATH_PREFIX_TMPXCMCNA9S': 'my_file'},
          'name': 'function2'}]
        '''
        result = FDLParser().parse_yaml('fdl.yaml')
        self.assertEqual(len(result), 2)
        for function in result:
            self.assertTrue(('name' in function) and ('env_vars' in function))
            if function['name'] == 'function1':
                self.assertEqual(len(function['env_vars'].items()), 8)
            elif function['name'] == 'function2':
                self.assertEqual(len(function['env_vars'].items()), 5)                


if __name__ == "__main__":
    unittest.main()



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with the class necessary to manage the
Cloudwatch Logs creation, deletion and configuration."""

from typing import Dict, List
from botocore.exceptions import ClientError
from scar.providers.aws.clients import BotoClient
from scar.exceptions import exception, ExistentLogGroupWarning, NotExistentLogGroupWarning
import scar.logger as logger


class CloudWatchLogsClient(BotoClient):
    """A low-level client representing Amazon CloudWatch Logs.
    DOC_URL: https://boto3.readthedocs.io/en/latest/reference/services/logs.html"""

    # Parameter used by the parent to create the appropriate boto3 client
    _BOTO_CLIENT_NAME = 'logs'

    @exception(logger)
    def get_log_events(self, **kwargs: Dict) -> List:
        """Lists log events from the specified log group."""
        log_events = []
        logs_info = self.client.filter_log_events(**kwargs)
        log_events.extend(logs_info.get('events', []))
        if 'nextToken' in logs_info:
            kwargs['nextToken'] = logs_info['nextToken']
            log_events.extend(self.get_log_events(**kwargs))
        return log_events

    @exception(logger)
    def create_log_group(self, **kwargs: Dict) -> Dict:
        """Creates a log group with the specified name."""
        try:
            return self.client.create_log_group(**kwargs)
        except ClientError as cerr:
            if cerr.response['Error']['Code'] == 'ResourceAlreadyExistsException':
                raise ExistentLogGroupWarning(logGroupName=kwargs['logGroupName'])
            raise cerr

    @exception(logger)
    def set_log_retention_policy(self, **kwargs: Dict) -> Dict:
        """Sets the retention of the specified log group."""
        return self.client.put_retention_policy(**kwargs)

    @exception(logger)
    def delete_log_group(self, log_group_name: str) -> Dict:
        """Deletes the specified log group and permanently deletes
        all the archived log events associated with the log group."""
        try:
            return self.client.delete_log_group(logGroupName=log_group_name)
        except ClientError as cerr:
            if cerr.response['Error']['Code'] == 'ResourceNotFoundException':
                raise NotExistentLogGroupWarning(logGroupName=log_group_name)
            raise cerr



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse


def create_function_definition_parser():
    function_definition_parser = argparse.ArgumentParser(add_help=False)
    function_definition_parser.add_argument("-d", "--description",
                                            help="Lambda function description.")
    function_definition_parser.add_argument("-e", "--environment",
                                            action='append',
                                            help=("Pass environment variable to the container "
                                                  "(VAR=val). Can be defined multiple times."))
    function_definition_parser.add_argument("-le", "--lambda-environment",
                                            action='append',
                                            help=("Pass environment variable to the lambda "
                                                  "function (VAR=val). Can be defined multiple "
                                                  "times."))
    function_definition_parser.add_argument("-m", "--memory",
                                            type=int,
                                            help=("Lambda function memory in megabytes. "
                                                  "Range from 128 to 3008 in increments of 64"))
    function_definition_parser.add_argument("-t", "--timeout",
                                            type=int,
                                            help=("Lambda function maximum execution "
                                                  "time in seconds. Max 900."))
    function_definition_parser.add_argument("-tt", "--timeout-threshold",
                                            type=int,
                                            help=("Extra time used to postprocess the data. "
                                                  "This time is extracted from the total "
                                                  "time of the lambda function."))
    function_definition_parser.add_argument("-ll", "--log-level",
                                            help=("Set the log level of the lambda function. "
                                                  "Accepted values are: "
                                                  "'CRITICAL','ERROR','WARNING','INFO','DEBUG'"))
    function_definition_parser.add_argument("-l", "--layers",
                                            action='append',
                                            help=("Pass layers ARNs to the lambda function. "
                                                  "Can be defined multiple times."))
    function_definition_parser.add_argument("-ib", "--input-bucket",
                                            help=("Bucket name where the input files "
                                                  "will be stored."))
    function_definition_parser.add_argument("-ob", "--output-bucket",
                                            help=("Bucket name where the output files are saved."))
    function_definition_parser.add_argument("-em", "--execution-mode",
                                            help=("Specifies the execution mode of the job. "
                                                  "It can be 'lambda', 'lambda-batch' or 'batch'"))
    function_definition_parser.add_argument("-r", "--iam-role",
                                            help=("IAM role used in the management of "
                                                  "the functions"))
    function_definition_parser.add_argument("-sv", "--supervisor-version",
                                            help=("FaaS Supervisor version. "
                                                  "Can be a tag or 'latest'."))
    # Batch (job definition) options
    function_definition_parser.add_argument("-bm", "--batch-memory",
                                            help="Batch job memory in megabytes")
    function_definition_parser.add_argument("-bc", "--batch-vcpus",
                                            help=("Number of vCPUs reserved for the "
                                                  "Batch container"))
    function_definition_parser.add_argument("-g", "--enable-gpu",
                                            help=("Reserve one physical GPU for the Batch "
                                                  "container (if it's available in the "
                                                  "compute environment)"),
                                            action="store_true")
    return function_definition_parser


def create_exec_parser():
    exec_parser = argparse.ArgumentParser(add_help=False)
    exec_parser.add_argument("-a", "--asynchronous",
                             help="Launch an asynchronous function.",
                             action="store_true")
    exec_parser.add_argument("-o", "--output-file",
                             help="Save output as a file")
    return exec_parser


def create_output_parser():
    output_parser = argparse.ArgumentParser(add_help=False)
    output_parser.add_argument("-j", "--json",
                               help="Return data in JSON format",
                               action="store_true")
    output_parser.add_argument("-v", "--verbose",
                               help="Show the complete aws output in json format",
                               action="store_true")
    return output_parser


def create_profile_parser():
    profile_parser = argparse.ArgumentParser(add_help=False)
    profile_parser.add_argument("-pf", "--profile",
                                help="AWS profile to use")
    return profile_parser


def create_storage_parser():
    storage_parser = argparse.ArgumentParser(add_help=False)
    storage_parser.add_argument("-b", "--bucket",
                                help="Bucket to use as storage",
                                required=True)
    storage_parser.add_argument("-p", "--path",
                                help="Path of the file or folder",
                                required=True)
    return storage_parser



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import subprocess
import unittest

class AwsTest(unittest.TestCase):
    
    scar_base_path = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
    scar_bin = ["scar"]
    
    def tearDown(self):
        self.execute_command(self.get_cmd(["rm","-a"]))
    
    def execute_command(self, cmd):
        return subprocess.check_output(cmd).decode("utf-8")
    
    def get_cmd(self, extra_args):
        return self.scar_bin + extra_args
    
    def create_function(self, function_name):
        cmd = self.get_cmd(["init","-n", function_name, "-i", "centos:7"])
        cmd_out = self.execute_command(cmd)
        self.assertTrue("Packing udocker files" in cmd_out)
        self.assertTrue("Creating function package." in cmd_out)
        self.assertTrue("Function '{0}' successfully created.".format(function_name) in cmd_out)
        self.assertTrue("Log group '/aws/lambda/{0}' successfully created.".format(function_name) in cmd_out)
     
    def test_empty_ls_table(self):
        cmd = self.get_cmd(["ls"])
        cmd_out = self.execute_command(cmd)
        self.assertEqual(cmd_out, 'NAME    MEMORY    TIME    IMAGE_ID    API_URL\n------  --------  ------  ----------  ---------\n')
          
        cmd = self.get_cmd(["ls", "-j"])
        cmd_out = self.execute_command(cmd)
        self.assertEqual(json.loads(cmd_out), {"Functions": []})
          
        cmd = self.get_cmd(["ls", "-v"])
        cmd_out = self.execute_command(cmd)
        self.assertEqual(json.loads(cmd_out), {"Functions": []})                 
                  
    def test_init_ls_run_rm_function(self):
        func_name = "scar-test-init-ls-run-rm"
        self.create_function(func_name)
         
        cmd = self.get_cmd(["ls"])
        cmd_out = self.execute_command(cmd)
        self.assertTrue("NAME                        MEMORY    TIME  IMAGE_ID    API_URL" in cmd_out)
        self.assertTrue("------------------------  --------  ------  ----------  ---------" in cmd_out)
        self.assertTrue("{0}       512     300  centos:7    -".format(func_name) in cmd_out)        
        
        cmd = self.get_cmd(["run", "-n", func_name])
        cmd_out = self.execute_command(cmd)
        self.assertTrue("Request Id:" in cmd_out)
        self.assertTrue("Log Group Name: /aws/lambda/{0}".format(func_name) in cmd_out)
        self.assertTrue("Log Stream Name:" in cmd_out)
                 
        cmd = self.get_cmd(["rm","-n", func_name])
        cmd_out = self.execute_command(cmd)
        self.assertTrue("Log group '/aws/lambda/{0}' successfully deleted".format(func_name) in cmd_out)
        self.assertTrue("Function '{0}' successfully deleted".format(func_name) in cmd_out)
         
    def test_init_ls_rm_function_json(self):
        func_name = "scar-test-init-ls-rm-json"
        self.create_function(func_name)
        
        cmd = self.get_cmd(["ls", "-j"])
        cmd_out = self.execute_command(cmd)
        self.assertEqual(json.loads(cmd_out),
                         {"Functions": [{"Name": func_name,
                                         "Memory": 512,
                                         "Timeout": 300,
                                         "Image_id": "centos:7",
                                         "Api_gateway": "-"}]
                         })
          
        cmd = self.get_cmd(["rm","-n", func_name, "-j"])
        cmd_out = self.execute_command(cmd)
        cw_out, lambda_out, _ = cmd_out.split("\n")
        self.assertTrue(json.loads(cw_out)['CloudWatchOutput']['HTTPStatusCode'], 200)
        self.assertTrue(json.loads(lambda_out)['LambdaOutput']['HTTPStatusCode'], 204)
         
    def test_init_ls_rm_simple_function_verbose(self):
        func_name = "scar-test-init-ls-rm-verbose"
        self.create_function(func_name)
        cmd = self.scar_bin + ["ls", "-v"]
        cmd_out = self.execute_command(cmd)
        output = json.loads(cmd_out)
        self.assertTrue(output['Functions'][0]['FunctionName'], "scar-func-test")
        self.assertTrue(output['Functions'][0]['Handler'], "scar-func-test.lambda_handler")

        cmd = self.scar_bin + ["rm","-n", func_name, "-v"]
        cmd_out = self.execute_command(cmd)
        cw_out,lambda_out,_ = cmd_out.split("\n")
        self.assertTrue(json.loads(cw_out)['CloudWatchOutput']['ResponseMetadata']['HTTPStatusCode'], 200)
        self.assertTrue(json.loads(lambda_out)['LambdaOutput']['ResponseMetadata']['HTTPStatusCode'], 204)
        
    def test_init_ls_rm_several_functions(self):
        function_names = ["scar-test-init-ls-rm-mult", "scar-test-init-ls-rm-mult-1", "scar-test-init-ls-rm-mult-2"]
        for function_name in function_names:
            self.create_function(function_name)

        cmd = self.get_cmd(["ls"])
        cmd_out = self.execute_command(cmd)
        self.assertTrue("NAME                           MEMORY    TIME  IMAGE_ID    API_URL" in cmd_out)
        self.assertTrue("---------------------------  --------  ------  ----------  ---------" in cmd_out)
        self.assertTrue("{0}         512     300  centos:7    -".format(function_names[0]) in cmd_out)
        self.assertTrue("{0}       512     300  centos:7    -".format(function_names[1]) in cmd_out)
        self.assertTrue("{0}       512     300  centos:7    -".format(function_names[2]) in cmd_out)

        cmd = self.get_cmd(["rm","-a"])
        cmd_out = self.execute_command(cmd)
        for function_name in function_names:
            self.assertTrue("Log group '/aws/lambda/{0}' successfully deleted".format(function_name) in cmd_out)
            self.assertTrue("Function '{0}' successfully deleted".format(function_name) in cmd_out)

if __name__ == '__main__':
    unittest.main()
    